// Copyright (c) Redux Robotics and other contributors.
// This is open source and can be modified and shared under the 3-clause BSD license.

package com.reduxrobotics.sensors.canandcolor;

import com.reduxrobotics.sensors.canandcolor.wpistruct.CanandcolorStatusStruct;

import edu.wpi.first.util.struct.StructSerializable;

/**
 * Container record class representing a {@link Canandcolor}'s status.
 * 
 * @param activeFaults the active faults of the device.
 * @param stickyFaults the sticky faults of the device.
 * @param temperature the onboard temperature of the device.
 */
public record CanandcolorStatus(CanandcolorFaults activeFaults, CanandcolorFaults stickyFaults, double temperature) implements StructSerializable {
    /** This field is necessary for WPILib structs to work around Java type erasure. */
    public static final CanandcolorStatusStruct struct = new CanandcolorStatusStruct();
    /**
     * Returns a default Status record for frame initializion
     * @return invalid Status record
     */
    public static CanandcolorStatus invalid() {
        return new CanandcolorStatus(new CanandcolorFaults(0, false), new CanandcolorFaults(0, false), 0);
    }

    /**
     * Deserializes a Status object from a byte array.
     * @param v array to deserialize from
     * @return new Status
     */
    public static CanandcolorStatus fromByteArray(byte[] v) {
        return new CanandcolorStatus( 
            new CanandcolorFaults(v[0], true),  // active faults
            new CanandcolorFaults(v[1], true),   // sticky faults
            (((int) v[0]) + ((int) v[1]) << 8) / 256.0 //temperature
        );
    }

    /**
     * Deserializes a Status object from a long.
     * @param data 64-bit long data
     * @return new Status
     */
    public static CanandcolorStatus fromLong(long data) {
        return new CanandcolorStatus(
            new CanandcolorFaults(CanandcolorDetails.Msg.extractStatus_Faults(data)),
            new CanandcolorFaults(CanandcolorDetails.Msg.extractStatus_StickyFaults(data)),
            CanandcolorDetails.Msg.extractStatus_Temperature(data) / 256.0
        );
    }

}