// Copyright (c) Redux Robotics and other contributors.
// This is open source and can be modified and shared under the 3-clause BSD license.

package com.reduxrobotics.sensors.canandcolor;

/**
 * Abstract base class for simple configuration of digout slot logic.
 * <h2>Running into errors trying to touch this class? </h2>
 * <p><b>Go look at the Javadoc for {@link RGBDigoutConfig} and {@link HSVDigoutConfig} for usage examples
 * with {@link DigoutChannel#configureSlots(DigoutConfig)}!</b></p>
 */
public sealed abstract class DigoutConfig permits HSVDigoutConfig, RGBDigoutConfig {
    DigoutConfig() {}

    double proximityHigh = 1;
    double proximityLow = 0;
    double proximityDebounce = 0;
    double redOrHueHigh = 1;
    double redOrHueLow = 0;
    double greenOrSatHigh = 1;
    double greenOrSatLow = 0;
    double blueOrValueHigh = 1;
    double blueOrValueLow = 0;
    double colorDebounce = 0;

    /**
     * Returns the digout chains used in {@link DigoutChannel#configureSlotsAdvanced(DigoutChain... chains)}
     * @return digout chain array
     */
    DigoutChain[] getDigoutChains() {
        DataSource redOrHue = isHSV() ? DataSource.kHue : DataSource.kRed;
        DataSource greenOrSat = isHSV() ? DataSource.kGreen : DataSource.kGreen;
        DataSource blueOrValue = isHSV() ? DataSource.kValue : DataSource.kBlue;

        DigoutChain[] ret = {
            DigoutChain.start()
                .lessThanEqualTo(DataSource.kProximity, proximityHigh) // slot 0
                .and()
                .greaterThanEqualTo(DataSource.kProximity, proximityLow) // slot 1
            .finish(),
            DigoutChain.start().prevChainTrueFor(proximityDebounce).finish(), // slot 2
            DigoutChain.start()
                .lessThanEqualTo(redOrHue, redOrHueHigh) // slot 3
                .join((redOrHueHigh < redOrHueLow && isHSV()) ? NextSlotAction.kOrWithNextSlot : NextSlotAction.kAndWithNextSlot)
                .greaterThanEqualTo(redOrHue, redOrHueLow) // slot 4
                .and()
                .lessThanEqualTo(greenOrSat, greenOrSatHigh) // slot 5
                .and()
                .greaterThanEqualTo(greenOrSat, greenOrSatLow) // slot 6
                .and()
                .lessThanEqualTo(blueOrValue, blueOrValueHigh) // slot 7
                .and()
                .greaterThanEqualTo(blueOrValue, blueOrValueLow) // slot 8
            .finish(),
            // slot 9
            DigoutChain.start().prevChainTrueFor(colorDebounce).finish()
        };
        return ret;
    }

    abstract boolean isHSV();

}
